
#include "quakedef.h"
#include "d3d_quake.h"
#include "d3d_hlsl.h"
#include "winquake.h"
#include "resource.h"

// keep these global as we'll want to use them in a few places
char *vs_version;
char *ps_version;


void D3D_MakeVertexDeclaration (int stream, int numverts, int numcolor, int numst, LPDIRECT3DVERTEXDECLARATION9 *d3d_VD)
{
	// set up the vertex element struct array
	D3DVERTEXELEMENT9 *ve = (D3DVERTEXELEMENT9 *) Heap_QMalloc (sizeof (D3DVERTEXELEMENT9) * (numverts + numcolor + numst + 1));

	// current item
	D3DVERTEXELEMENT9 *currentve = ve;
	int ofs = 0;

	for (int i = 0; i < numverts; i++, currentve++)
	{
		// fill 'er in
		currentve->Stream = stream;
		currentve->Offset = ofs;
		currentve->Type = D3DDECLTYPE_FLOAT3;
		currentve->Method = D3DDECLMETHOD_DEFAULT;
		currentve->Usage = D3DDECLUSAGE_POSITION;
		currentve->UsageIndex = i;

		// next offset
		ofs += 12;
	}

	for (int i = 0; i < numcolor; i++, currentve++)
	{
		// fill 'er in
		currentve->Stream = stream;
		currentve->Offset = ofs;
		currentve->Type = D3DDECLTYPE_D3DCOLOR;
		currentve->Method = D3DDECLMETHOD_DEFAULT;
		currentve->Usage = D3DDECLUSAGE_COLOR;
		currentve->UsageIndex = i;

		// next offset
		ofs += 4;
	}

	for (int i = 0; i < numst; i++, currentve++)
	{
		// fill 'er in
		currentve->Stream = stream;
		currentve->Offset = ofs;
		currentve->Type = D3DDECLTYPE_FLOAT2;
		currentve->Method = D3DDECLMETHOD_DEFAULT;
		currentve->Usage = D3DDECLUSAGE_TEXCOORD;
		currentve->UsageIndex = i;

		// next offset
		ofs += 8;
	}

	// terminate it (D3DDECL_END)
	currentve->Stream = 0xff;
	currentve->Offset = 0;
	currentve->Type = D3DDECLTYPE_UNUSED;
	currentve->Method = 0;
	currentve->Usage = 0;
	currentve->UsageIndex = 0;

	// create the declaration
	d3d_Device->CreateVertexDeclaration (ve, d3d_VD);

	// done
	Heap_QFreeFull (ve);
}


// effects
LPD3DXEFFECT d3d_InstancedBrushFX = NULL;
LPD3DXEFFECT d3d_BrushFX = NULL;
LPD3DXEFFECT d3d_LiquidFX = NULL;
LPD3DXEFFECT d3d_AliasFX = NULL;
LPD3DXEFFECT d3d_Flat2DFX = NULL;
LPD3DXEFFECT d3d_SpriteFX = NULL;
LPD3DXEFFECT d3d_ParticleFX = NULL;
LPD3DXEFFECT d3d_UnderwaterFX = NULL;

// vertex declarations
LPDIRECT3DVERTEXDECLARATION9 d3d_AliasVertexDeclaration = NULL;
LPDIRECT3DVERTEXDECLARATION9 d3d_ParticleVertexDeclaration = NULL;
LPDIRECT3DVERTEXDECLARATION9 d3d_V3ST2Declaration = NULL;
LPDIRECT3DVERTEXDECLARATION9 d3d_V3ST4Declaration = NULL;

void D3D_LoadEffect (char *name, int resourceid, LPD3DXEFFECT *eff)
{
	// because there are so many places this can go wrong we'll use exception handling on it rather
	// than creating a mess of error-checking.
	try
	{
		if (!name)
		{
			Sys_Error ("File name alternate for shader creation not specified!");
			return;
		}

		// first off we try to create the shader from a file, so that we'll be able to let
		// users play with their own shaders if they like.  if that fails we'll create it from
		// a resource instead to use a pre-canned shader.  (TO DO!)

		// load the resource - note - we don't use D3DXCreateEffectFromResource as importing an RCDATA resource
		// from a text file in Visual Studio doesn't NULL terminate the file, causing it to blow up.
		HRSRC hResInfo = FindResource (NULL, MAKEINTRESOURCE (resourceid), RT_RCDATA);
		HGLOBAL hResData = LoadResource (NULL, hResInfo);
		char *EffectString = (char *) LockResource (hResData);

		LPD3DXBUFFER errbuf;

		// this madness is required because sometimes the byte alignment of an RCDATA resource
		// tacks on an extra character at the end.  I'm certain it's probably documented SOMEWHERE
		// that RCDATA should be WORD-aligned or whatever, but when it comes to MS documentation
		// I'm reminded of Arthur Dent and the council's bypass plans, so rather than waste my
		// time I'm just going to read backwards until I get what I want.  Sigh.
		int len = strlen (EffectString);

		for (; len; len--)
		{
			// check for end of technique
			if (EffectString[len] == '}')
			{
				len++;
				break;
			}
		}

		HRESULT hr = D3DXCreateEffect
		(
			d3d_Device,
			EffectString,
			len,
			NULL,
			NULL,
			D3DXSHADER_OPTIMIZATION_LEVEL3 | D3DXSHADER_SKIPVALIDATION | D3DXSHADER_PARTIALPRECISION,
			NULL,
			eff,
			&errbuf
		);

		// clean up
		UnlockResource (hResData);

		if (FAILED (hr))
		{
			char *errstr = (char *) errbuf->GetBufferPointer ();
			Sys_Error ("Error compiling %s\n%s", name, errstr);
			errbuf->Release ();
			return;
		}

		Con_Printf ("Created effect \"%s\" OK\n", name);
	}
	catch (...)
	{
	}
}


void D3D_InitHLSL (void)
{
	vs_version = (char *) D3DXGetVertexShaderProfile (d3d_Device);
	ps_version = (char *) D3DXGetPixelShaderProfile (d3d_Device);

	Con_Printf ("\n");

	if (vs_version)
		Con_Printf ("Vertex Shader Version: %s\n", vs_version);
	else
	{
		// note - we expect this to never happen as D3D will create vertex shaders in software
		// for devices that don't support them in hardware.
		Sys_Error ("Vertex Shaders Not Available\n");
		return;
	}

	if (ps_version)
		Con_Printf ("Pixel Shader Version: %s\n", ps_version);
	else
	{
		// note - we expect this to never happen as D3D will create vertex shaders in software
		// for devices that don't support them in hardware.
		Sys_Error ("Pixel Shaders Not Available\n");
		return;
	}

	Con_Printf ("\n");

	// load effects
	D3D_LoadEffect ("AliasModel.fx", IDR_ALIAS, &d3d_AliasFX);
	D3D_LoadEffect ("Brush.fx", IDR_BRUSH, &d3d_BrushFX);
	D3D_LoadEffect ("Flat2D.fx", IDR_FLAT2D, &d3d_Flat2DFX);
	D3D_LoadEffect ("InstancedBrush.fx", IDR_INSTANCEDBRUSH, &d3d_InstancedBrushFX);
	D3D_LoadEffect ("Liquid.fx", IDR_LIQUID, &d3d_LiquidFX);
	D3D_LoadEffect ("Particle.fx", IDR_PARTICLE, &d3d_ParticleFX);
	D3D_LoadEffect ("Sprite.fx", IDR_SPRITE, &d3d_SpriteFX);
	D3D_LoadEffect ("Underwater.fx", IDR_UNDERWATER, &d3d_UnderwaterFX);

	// set up vertex declarations
	D3D_MakeVertexDeclaration (0, 2, 1, 2, &d3d_AliasVertexDeclaration);
	D3D_MakeVertexDeclaration (0, 1, 1, 1, &d3d_ParticleVertexDeclaration);
	D3D_MakeVertexDeclaration (0, 1, 0, 1, &d3d_V3ST2Declaration);
	D3D_MakeVertexDeclaration (0, 1, 0, 2, &d3d_V3ST4Declaration);
}


void D3D_ShutdownHLSL (void)
{
	// effects
	SAFE_RELEASE (d3d_InstancedBrushFX);
	SAFE_RELEASE (d3d_BrushFX);
	SAFE_RELEASE (d3d_LiquidFX);
	SAFE_RELEASE (d3d_AliasFX);
	SAFE_RELEASE (d3d_Flat2DFX);
	SAFE_RELEASE (d3d_SpriteFX);
	SAFE_RELEASE (d3d_ParticleFX);
	SAFE_RELEASE (d3d_UnderwaterFX);

	// declarations
	SAFE_RELEASE (d3d_AliasVertexDeclaration);
	SAFE_RELEASE (d3d_ParticleVertexDeclaration);
	SAFE_RELEASE (d3d_V3ST2Declaration);
	SAFE_RELEASE (d3d_V3ST4Declaration);
}


